// This is a data store backend for the Citadel server which uses Berkeley DB.
//
// Copyright (c) 1987-2023 by the citadel.org team
//
// This program is open source software.  Use, duplication, or disclosure
// is subject to the terms of the GNU General Public License, version 3.

// Citadel will checkpoint the db at the end of every session, but only if
// the specified number of kilobytes has been written, or if the specified
// number of minutes has passed, since the last checkpoint.
#define MAX_CHECKPOINT_KBYTES	256
#define MAX_CHECKPOINT_MINUTES	15

#include "../../sysdep.h"
#include <sys/stat.h>
#include <stdio.h>
#include <zlib.h>
#include <db.h>
#include <assert.h>
#include <libcitadel.h>
#include "../../citserver.h"
#include "../../config.h"
#include "berkeley_db.h"

#if DB_VERSION_MAJOR < 18
#error Citadel requires Berkeley DB v18.0 or newer.  Please upgrade.
#endif

// Globals (these are used across multiple functions in *this* module, but not elsewhere)
static DB *bdb_table[MAXCDB];	// One DB handle for each Citadel database
static DB_ENV *bdb_env;		// The DB environment (global)


// These are items that we need to store "per thread" rather than "per session".
struct bdb_tsd {
	DB_TXN *tid;            // Transaction handle
	DBC *cursors[MAXCDB];   // Cursors, for traversals...
	DBT dbkey[MAXCDB];
	DBT dbdata[MAXCDB];
};


pthread_key_t bdb_thread_key;
#define TSD bdb_get_tsd()


// Some functions in this backend need to store some per-thread data.
// This returns the pointer to the current thread's per-thread data block, creating it if necessary.
// (This will also work in a non-threaded program; it will return the same pointer every time.)
struct bdb_tsd *bdb_get_tsd(void) {

        struct bdb_tsd *c = (struct bdb_tsd *) pthread_getspecific(bdb_thread_key) ;
	if (c != NULL) {
		return(c);		// Got it.
	}

	// If there's no TSD for this thread, it must be a new thread.  Create our TSD region.
	c = (struct bdb_tsd *) malloc(sizeof(struct bdb_tsd));
	memset(c, 0, sizeof(struct bdb_tsd));
	pthread_setspecific(bdb_thread_key, (const void *) c);
	return(c);
}


// Called by other functions in this module to GTFO quickly if we need to.  Not part of the backend API.
void bdb_abort(void) {
	syslog(LOG_DEBUG, "bdb: citserver is stopping in order to prevent data loss. uid=%d gid=%d euid=%d egid=%d",
		getuid(), getgid(), geteuid(), getegid()
	);
	raise(SIGABRT);		// This will exit in a way that can produce a core dump if needed.
	exit(CTDLEXIT_DB);	// Exit if the signal failed to end the program.
}


// Verbose logging callback for Berkeley DB.  Not part of the backend API.
void bdb_verbose_log(const DB_ENV *bdb_env, const char *msg, const char *foo) {
	if (!IsEmptyStr(msg)) {
		syslog(LOG_DEBUG, "bdb: %s %s", msg, foo);
	}
}


// Verbose error logging callback for Berkeley DB.  Not part of the backend API.
void bdb_verbose_err(const DB_ENV *bdb_env, const char *errpfx, const char *msg) {
	syslog(LOG_ERR, "bdb: %s", msg);
}


// Wrapper for txn_abort() that logs/aborts on error.  Not part of the backend API.
static void bdb_txabort(DB_TXN *tid) {
	int ret;

	ret = tid->abort(tid);

	if (ret) {
		syslog(LOG_ERR, "bdb: txn_abort: %s", db_strerror(ret));
		bdb_abort();
	}
}


// Wrapper for txn_commit() that logs/aborts on error.  Not part of the backend API.
static void bdb_txcommit(DB_TXN *tid) {
	int ret;

	ret = tid->commit(tid, 0);

	if (ret) {
		syslog(LOG_ERR, "bdb: txn_commit: %s", db_strerror(ret));
		bdb_abort();
	}
}


// Panic callback for Berkeley DB.  Not part of the backend API.
static void bdb_dbpanic(DB_ENV *env, int errval) {
	syslog(LOG_ERR, "bdb: PANIC: %s", db_strerror(errval));
	bdb_abort();
}


// Close a cursor -- not part of the backend API.
static void bdb_cclose(DBC *cursor) {
	int ret;

	if ((ret = cursor->c_close(cursor))) {
		syslog(LOG_ERR, "bdb: c_close: %s", db_strerror(ret));
		bdb_abort();
	}
}


// Convenience function to abort if a cursor is still open when we didn't expect it to be.
// This is not part of the backend API.
static void bdb_bailIfCursor(DBC **cursors, const char *msg) {
	int i;

	for (i = 0; i < MAXCDB; i++)
		if (cursors[i] != NULL) {
			syslog(LOG_ERR, "bdb: cursor still in progress on cdb %02x: %s", i, msg);
			bdb_abort();
		}
}


void bdb_check_handles(void) {
	bdb_bailIfCursor(TSD->cursors, "in check_handles");

	if (TSD->tid != NULL) {
		syslog(LOG_ERR, "bdb: transaction still in progress!");
		bdb_abort();
	}
}


// Transaction-based stuff.  I'm writing this as I bake cookies...
void bdb_begin_transaction(void) {
	int ret;
	bdb_bailIfCursor(TSD->cursors, "can't begin transaction during r/o cursor");

	if (TSD->tid != NULL) {
		syslog(LOG_ERR, "bdb: bdb_begin_transaction: ERROR: nested transaction");
		bdb_abort();
	}

	ret = bdb_env->txn_begin(bdb_env, NULL, &TSD->tid, 0);
	if (ret) {
		syslog(LOG_ERR, "bdb: bdb_begin_transaction: %s", db_strerror(ret));
		bdb_abort();
	}
}


// ...and the cookies are cursed.
void bdb_end_transaction(void) {
	int i;

	for (i = 0; i < MAXCDB; i++) {
		if (TSD->cursors[i] != NULL) {
			syslog(LOG_WARNING, "bdb: bdb_end_transaction: WARNING: cursor %d still open at transaction end", i);
			bdb_cclose(TSD->cursors[i]);
			TSD->cursors[i] = NULL;
		}
	}

	if (TSD->tid == NULL) {
		syslog(LOG_ERR, "bdb: bdb_end_transaction: ERROR: bdb_txcommit(NULL) !!");
		bdb_abort();
	}
	else {
		bdb_txcommit(TSD->tid);
	}

	TSD->tid = NULL;
}


// Request a checkpoint of the database.  Called once per minute by the thread manager.
void bdb_checkpoint(void) {
	int ret;

	syslog(LOG_DEBUG, "bdb: -- checkpoint --");
	ret = bdb_env->txn_checkpoint(bdb_env, MAX_CHECKPOINT_KBYTES, MAX_CHECKPOINT_MINUTES, 0);

	if (ret != 0) {
		syslog(LOG_ERR, "bdb: bdb_checkpoint() txn_checkpoint: %s", db_strerror(ret));
		bdb_abort();
	}

	// After a successful checkpoint, we can cull the unused logs
	ret = bdb_env->log_set_config(bdb_env, DB_LOG_AUTO_REMOVE, 1);
	if (ret != 0) {
		syslog(LOG_ERR, "bdb: bdb_checkpoint() auto coll logs: %s", db_strerror(ret));
	}
}


// Open the various tables we'll be using.  Any table which does not exist should be created.  Note that we don't need a
// critical section here, because there aren't any active threads manipulating the database yet.
void bdb_open_databases(void) {
	int ret;
	int i;
	char dbfilename[32];
	u_int32_t flags = 0;
	int dbversion_major, dbversion_minor, dbversion_patch;

	syslog(LOG_DEBUG, "bdb: bdb_open_databases() starting");
	syslog(LOG_DEBUG, "bdb:    Linked zlib: %s", zlibVersion());
	syslog(LOG_DEBUG, "bdb: Compiled libdb: %s", DB_VERSION_STRING);
	syslog(LOG_DEBUG, "bdb:   Linked libdb: %s", db_version(&dbversion_major, &dbversion_minor, &dbversion_patch));

	// Create synthetic integer version numbers and compare them.
	// Never allow citserver to run with a libdb older then the one with which it was compiled.
	int compiled_db_version = ( (DB_VERSION_MAJOR * 1000000) + (DB_VERSION_MINOR * 1000) + (DB_VERSION_PATCH) );
	int linked_db_version = ( (dbversion_major * 1000000) + (dbversion_minor * 1000) + (dbversion_patch) );
	if (compiled_db_version > linked_db_version) {
		syslog(LOG_ERR, "bdb: citserver is running with a version of libdb older than the one with which it was compiled.");
		syslog(LOG_ERR, "bdb: This is an invalid configuration.  citserver will now exit to prevent data loss.");
		exit(CTDLEXIT_DB);
	}

	syslog(LOG_DEBUG, "bdb: Setting up DB environment");
	ret = db_env_create(&bdb_env, 0);
	if (ret) {
		syslog(LOG_ERR, "bdb: db_env_create: %s", db_strerror(ret));
		syslog(LOG_ERR, "bdb: exit code %d", ret);
		exit(CTDLEXIT_DB);
	}
	bdb_env->set_errpfx(bdb_env, "citserver");
	bdb_env->set_paniccall(bdb_env, bdb_dbpanic);
	bdb_env->set_errcall(bdb_env, bdb_verbose_err);
	bdb_env->set_msgcall(bdb_env, bdb_verbose_log);
	bdb_env->set_verbose(bdb_env, DB_VERB_DEADLOCK, 1);
	bdb_env->set_verbose(bdb_env, DB_VERB_RECOVERY, 1);

	flags = DB_CREATE | DB_INIT_LOCK | DB_INIT_LOG | DB_INIT_MPOOL | DB_INIT_TXN | DB_RECOVER | DB_THREAD ;
	syslog(LOG_DEBUG, "bdb: bdb_env->open(bdb_env, %s, %d, 0)", ctdl_db_dir, flags);
	ret = bdb_env->open(bdb_env, ctdl_db_dir, flags, 0);				// try opening the database cleanly
	if (ret) {
		syslog(LOG_ERR, "bdb: bdb_env->open: %s: %s", ctdl_db_dir, db_strerror(ret));
		bdb_env->close(bdb_env, 0);
		syslog(LOG_ERR, "bdb: exit code %d", ret);
		exit(CTDLEXIT_DB);
	}

	for (i = 0; i < MAXCDB; ++i) {
		ret = db_create(&bdb_table[i], bdb_env, 0);				// Create a database handle
		if (ret) {
			syslog(LOG_ERR, "bdb: db_create: %s", db_strerror(ret));
			syslog(LOG_ERR, "bdb: exit code %d", ret);
			exit(CTDLEXIT_DB);
		}

		snprintf(dbfilename, sizeof dbfilename, "cdb.%02x", i);			// table names by number
		ret = bdb_table[i]->open(bdb_table[i], NULL, dbfilename, NULL, DB_BTREE, DB_CREATE | DB_AUTO_COMMIT, 0600);
		if (ret) {
			syslog(LOG_ERR, "bdb: db_open[%02x]: %s", i, db_strerror(ret));
			if (ret == ENOMEM) {
				syslog(LOG_ERR, "bdb: You may need to tune your database; please check http://www.citadel.org for more information.");
			}
			syslog(LOG_ERR, "bdb: exit code %d", ret);
			exit(CTDLEXIT_DB);
		}
	}
}


// Close all of the db database files we've opened.  This can be done in a loop, since it's just a bunch of closes.
void bdb_close_databases(void) {
	int i;
	int ret;

	static int closing = 0;
	while (closing == 1) {
		syslog(LOG_INFO, "bdb: already closing");
	}
	closing = 1;

	syslog(LOG_INFO, "bdb: performing final checkpoint");
	if ((ret = bdb_env->txn_checkpoint(bdb_env, 0, 0, 0))) {
		syslog(LOG_ERR, "bdb: txn_checkpoint: %s", db_strerror(ret));
	}

	syslog(LOG_INFO, "bdb: flushing the database logs");
	if ((ret = bdb_env->log_flush(bdb_env, NULL))) {
		syslog(LOG_ERR, "bdb: log_flush: %s", db_strerror(ret));
	}

	// close the tables
	syslog(LOG_INFO, "bdb: closing databases");
	for (i = 0; i < MAXCDB; ++i) {
		ret = bdb_table[i]->close(bdb_table[i], 0);
		if (ret) {
			syslog(LOG_ERR, "bdb: db_close: %s", db_strerror(ret));
		}
	}

	// Close the handle.
	syslog(LOG_INFO, "bdb: closing environment");
	ret = bdb_env->close(bdb_env, DB_FORCESYNC);
	if (ret) {
		syslog(LOG_ERR, "bdb: DBENV->close: %s", db_strerror(ret));
	}

	syslog(LOG_INFO, "bdb: shutdown completed");
}


// Decompress a DBT data block if it was compressed on disk.
void bdb_decompress_if_necessary(DBT *d) {
	static int magic = COMPRESS_MAGIC;

	if ((d == NULL) || (d->data == NULL) || (d->size < sizeof(magic)) || (memcmp(d->data, &magic, sizeof(magic)))) {
		return;
	}

	// At this point we know we're looking at a compressed item.

	struct CtdlCompressHeader zheader;
	char *uncompressed_data;
	char *compressed_data;
	uLongf destLen, sourceLen;
	size_t cplen;

	memset(&zheader, 0, sizeof(struct CtdlCompressHeader));
	cplen = sizeof(struct CtdlCompressHeader);
	if (sizeof(struct CtdlCompressHeader) > d->size) {
		cplen = d->size;
	}
	memcpy(&zheader, d->data, cplen);

	compressed_data = d->data;
	compressed_data += sizeof(struct CtdlCompressHeader);

	sourceLen = (uLongf) zheader.compressed_len;
	destLen = (uLongf) zheader.uncompressed_len;
	uncompressed_data = malloc(zheader.uncompressed_len);

	if (uncompress((Bytef *) uncompressed_data, (uLongf *) &destLen, (const Bytef *) compressed_data, (uLong) sourceLen) != Z_OK) {
		syslog(LOG_ERR, "bdb: uncompress() error");
		bdb_abort();
	}

	free(d->data);
	d->size = (size_t) destLen;
	d->data = uncompressed_data;
}


// Store a piece of data.  Returns 0 if the operation was successful.  If a
// key already exists it should be overwritten.
int bdb_store(int cdb, const void *ckey, int ckeylen, void *cdata, int cdatalen) {

	DBT dkey, ddata;
	DB_TXN *tid = NULL;
	int ret = 0;
	struct CtdlCompressHeader zheader;
	char *compressed_data = NULL;
	int compressing = 0;
	size_t buffer_len = 0;
	uLongf destLen = 0;
	int existing_txn = 0;			// set to nonzero if we are already inside a transaction

	memset(&dkey, 0, sizeof(DBT));
	memset(&ddata, 0, sizeof(DBT));
	dkey.size = ckeylen;
	dkey.data = (void *) ckey;
	ddata.size = cdatalen;
	ddata.data = cdata;

	// "visit" records are numerous and have big, mostly-empty string buffers in them.
	// If we compress these we can get them down to 1% of their size most of the time.
	if (cdb == CDB_VISIT) {
		compressing = 1;
		zheader.magic = COMPRESS_MAGIC;
		zheader.uncompressed_len = cdatalen;
		buffer_len = ((cdatalen * 101) / 100) + 100 + sizeof(struct CtdlCompressHeader);
		destLen = (uLongf) buffer_len;
		compressed_data = malloc(buffer_len);
		if (compress2((Bytef *) (compressed_data + sizeof(struct CtdlCompressHeader)), &destLen, (Bytef *) cdata, (uLongf) cdatalen, 1) != Z_OK) {
			syslog(LOG_ERR, "bdb: compress2() error");
			bdb_abort();
		}
		zheader.compressed_len = (size_t) destLen;
		memcpy(compressed_data, &zheader, sizeof(struct CtdlCompressHeader));
		ddata.size = (size_t) (sizeof(struct CtdlCompressHeader) + zheader.compressed_len);
		ddata.data = compressed_data;
	}

	if (TSD->tid != NULL) {
		existing_txn = 1;
	}

	if (!existing_txn) {				// If we're not already inside a transaction,
		bdb_begin_transaction();		// create our own for this operation.
	}

	do {
		ret = bdb_table[cdb]->put(bdb_table[cdb],	// db
			    	TSD->tid,			// transaction ID
			    	&dkey,			// key
			    	&ddata,			// data
			    	0				// flags
		);
		if ((ret != 0) && (ret != DB_LOCK_DEADLOCK)) {
			syslog(LOG_ERR, "bdb: bdb_store(%02x): error %d: %s", cdb, ret, db_strerror(ret));
			bdb_abort();
		}
		if (ret == DB_LOCK_DEADLOCK) {
			syslog(LOG_DEBUG, "bdb: bdb_store(%02x): would deadlock, trying again", cdb);
		}
	} while (ret == DB_LOCK_DEADLOCK);

	if (!existing_txn) {
		bdb_end_transaction();
	}

	if (compressing) {
		free(compressed_data);
	}

	return ret;
}


// Delete a piece of data.  Returns 0 if the operation was successful.
int bdb_delete(int cdb, void *key, int keylen) {
	DBT dkey;
	int ret;
	int existing_txn = 0;				// set to nonzero if we are already inside a transaction

	memset(&dkey, 0, sizeof dkey);
	dkey.size = keylen;
	dkey.data = key;

	if (TSD->tid != NULL) {
		existing_txn = 1;
	}

	if (!existing_txn) {				// If we're not already inside a transaction,
		bdb_begin_transaction();		// create our own for this operation.
	}

	ret = bdb_table[cdb]->del(bdb_table[cdb], TSD->tid, &dkey, 0);
	if (ret) {
		if (ret != DB_NOTFOUND) {
			syslog(LOG_ERR, "bdb: bdb_delete(%02x): %s", cdb, db_strerror(ret));
			bdb_abort();
		}
	}

	if (!existing_txn) {
		bdb_end_transaction();			// Only end the transaction if we began it.
	}

	return ret;
}


// Fetch a piece of data.  Returns a "struct cdbdata"
// If the item is not found, the pointer will be NULL.
struct cdbdata bdb_fetch(int cdb, const void *key, int keylen) {

	struct cdbdata returned_data;
	memset(&returned_data, 0, sizeof(struct cdbdata));

	if (keylen == 0) {		// key length zero is impossible
		return(returned_data);
	}

	DBT dkey;
	int ret;

	memset(&dkey, 0, sizeof(DBT));
	dkey.size = keylen;
	dkey.data = (void *) key;

	TSD->dbdata[cdb].flags = DB_DBT_REALLOC;

	do {
		ret = bdb_table[cdb]->get(bdb_table[cdb], TSD->tid, &dkey, &TSD->dbdata[cdb], 0);
		if ((ret != 0) && (ret != DB_NOTFOUND) && (ret != DB_LOCK_DEADLOCK)) {
			syslog(LOG_ERR, "bdb: bdb_fetch(%d): error %d: %s", cdb, ret, db_strerror(ret));
			bdb_abort();
		}
	} while (ret == DB_LOCK_DEADLOCK);

	if (ret == 0) {
		bdb_decompress_if_necessary(&TSD->dbdata[cdb]);
		returned_data.len = TSD->dbdata[cdb].size;
		returned_data.ptr = TSD->dbdata[cdb].data;
	}

	return(returned_data);
}


void bdb_close_cursor(int cdb) {
	if (TSD->cursors[cdb] != NULL) {
		bdb_cclose(TSD->cursors[cdb]);
	}

	TSD->cursors[cdb] = NULL;
}


// Prepare for a sequential search of an entire database.
// (There is guaranteed to be no more than one traversal in progress per thread at any given time.)
void bdb_rewind(int cdb) {
	int ret = 0;

	if (TSD->cursors[cdb] != NULL) {
		syslog(LOG_ERR, "bdb: bdb_rewind: must close cursor on database %d before reopening", cdb);
		bdb_abort();
		// bdb_cclose(TSD->cursors[cdb]);
	}

	// Now initialize the cursor
	ret = bdb_table[cdb]->cursor(bdb_table[cdb], TSD->tid, &TSD->cursors[cdb], 0);
	if (ret) {
		syslog(LOG_ERR, "bdb: bdb_rewind: db_cursor: %s", db_strerror(ret));
		bdb_abort();
	}
}


// Fetch the next item in a sequential search.  Returns a pointer to a 
// cdbdata structure, or NULL if we've hit the end.
struct cdbkeyval bdb_next_item(int cdb) {
	struct cdbkeyval kv;
	int ret = 0;

	memset(&kv, 0, sizeof(struct cdbkeyval));

	// reuse memory from the previous call.
	TSD->dbkey[cdb].flags = DB_DBT_REALLOC;
	TSD->dbdata[cdb].flags = DB_DBT_REALLOC;

	assert(TSD->cursors[cdb] != NULL);
	ret = TSD->cursors[cdb]->c_get(TSD->cursors[cdb], &TSD->dbkey[cdb], &TSD->dbdata[cdb], DB_NEXT);

	if (ret) {
		if (ret != DB_NOTFOUND) {
			syslog(LOG_ERR, "bdb: bdb_next_item(%d): %s", cdb, db_strerror(ret));
			bdb_abort();
		}
		bdb_close_cursor(cdb);
		return(kv);		// presumably, we are at the end
	}

	bdb_decompress_if_necessary(&TSD->dbdata[cdb]);

	kv.key.len = TSD->dbkey[cdb].size;
	kv.key.ptr = TSD->dbkey[cdb].data;
	kv.val.len = TSD->dbdata[cdb].size;
	kv.val.ptr = TSD->dbdata[cdb].data;
	return (kv);
}


// Truncate (delete every record)
void bdb_trunc(int cdb) {
	int ret;
	u_int32_t count;

	if (TSD->tid != NULL) {
		syslog(LOG_ERR, "bdb: bdb_trunc must not be called in a transaction.");
		bdb_abort();
	}

	bdb_begin_transaction();				// create our own transaction for this operation.
	ret = bdb_table[cdb]->truncate(bdb_table[cdb],		// db
				      NULL,			// transaction ID
				      &count,			// #rows deleted
				      0);			// flags
								//
	if (ret) {
		syslog(LOG_ERR, "bdb: bdb_truncate(%d): %s", cdb, db_strerror(ret));
		if (ret == ENOMEM) {
			syslog(LOG_ERR, "bdb: You may need to tune your database; please read http://www.citadel.org for more information.");
		}
		exit(CTDLEXIT_DB);
	}
	bdb_end_transaction();
}


// compact (defragment) the database, possibly returning space back to the underlying filesystem
void bdb_compact(void) {
	int ret;
	int i;

	syslog(LOG_DEBUG, "bdb: bdb_compact() started");
	for (i = 0; i < MAXCDB; i++) {
		syslog(LOG_DEBUG, "bdb: compacting database %d", i);
		ret = bdb_table[i]->compact(bdb_table[i], NULL, NULL, NULL, NULL, DB_FREE_SPACE, NULL);
		if (ret) {
			syslog(LOG_ERR, "bdb: compact: %s", db_strerror(ret));
		}
	}
	syslog(LOG_DEBUG, "bdb: bdb_compact() finished");
}


// periodically called for maintenance
void bdb_tick(void) {
	int ret;
	int rejected;

	ret = bdb_env->lock_detect(bdb_env, 0, DB_LOCK_DEFAULT, &rejected);
	if (ret) {
		syslog(LOG_ERR, "bdb: lock_detect: %s", db_strerror(ret));
	}
	else if (rejected) {
		syslog(LOG_DEBUG, "bdb: rejected lock %d", rejected);
	}
}


// Calling this function activates the Berkeley DB back end.
void bdb_init_backend(void) {

	// Assign the backend API stubs to the functions in this module.
	cdb_compact = bdb_compact;
	cdb_checkpoint = bdb_checkpoint;
	cdb_rewind = bdb_rewind;
	cdb_fetch = bdb_fetch;
	cdb_open_databases = bdb_open_databases;
	cdb_close_databases = bdb_close_databases;
	cdb_store = bdb_store;
	cdb_delete = bdb_delete;
	cdb_next_item = bdb_next_item;
	cdb_close_cursor = bdb_close_cursor;
	cdb_begin_transaction = bdb_begin_transaction;
	cdb_end_transaction = bdb_end_transaction;
	cdb_check_handles = bdb_check_handles;
	cdb_trunc = bdb_trunc;
	cdb_tick = bdb_tick;

	// Some functions in this backend need to store some per-thread data.
	// We create the key here, during module initialization.
	if (pthread_key_create(&bdb_thread_key, NULL) != 0) {
		syslog(LOG_ERR, "pthread_key_create() : %m");
		exit(CTDLEXIT_THREAD);
	}

	syslog(LOG_INFO, "db: initialized Berkeley DB backend");
}
