// This module handles fulltext indexing of the message base.
//
// Copyright (c) 2005-2023 by the citadel.org team
//
// This program is open source software.  Use, duplication, or disclosure
// is subject to the terms of the GNU General Public License, version 3.

#include "../../sysdep.h"
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <pwd.h>
#include <errno.h>
#include <sys/types.h>
#include <time.h>
#include <sys/wait.h>
#include <string.h>
#include <limits.h>
#include <libcitadel.h>
#include "../../citadel_defs.h"
#include "../../server.h"
#include "../../citserver.h"
#include "../../support.h"
#include "../../config.h"
#include "../../room_ops.h"
#include "../../database.h"
#include "../../msgbase.h"
#include "../../control.h"
#include "serv_fulltext.h"
#include "ft_wordbreaker.h"
#include "../../threads.h"
#include "../../context.h"
#include "../../ctdl_module.h"

// These can be global variables because only one indexer runs at a time.
Array *messages_to_be_indexed = NULL;
long highest_msg_already_indexed = 0;
long highest_msg_to_be_indexed = 0;


// Compare function
int longcmp(const void *rec1, const void *rec2) {
	long i1, i2;

	i1 = *(const long *)rec1;
	i2 = *(const long *)rec2;

	if (i1 > i2) return(1);
	if (i1 < i2) return(-1);
	return(0);
}


// Index or de-index a message.  (op == 1 to index, 0 to de-index)
void ft_index_message(long msgnum, int op) {
	int i, j;
	Array *tokens_in_this_message = NULL;
	struct cdbdata cdb_bucket;
	StrBuf *msgtext;
	char *txt;
	int tok;
	struct CtdlMessage *msg = NULL;

	if (msgnum == 0) return;

	msg = CtdlFetchMessage(msgnum, 1);
	if (msg == NULL) {
		// This is not necessarily an error condition; it could simply mean that the message was
		// deleted before it could be indexed.  This happens often when the load tester is running.
		syslog(LOG_ERR, "fulltext: ft_index_message() could not load msg %ld", msgnum);
		return;
	}

	if (!CM_IsEmpty(msg, eSuppressIdx)) {
		syslog(LOG_DEBUG, "fulltext: ft_index_message() excluded msg %ld", msgnum);
		CM_Free(msg);
		return;
	}

	syslog(LOG_DEBUG, "fulltext: ft_index_message() %s msg %ld", (op ? "adding" : "removing") , msgnum);

	// Output the message as text before indexing it, so we don't end up indexing a bunch of encoded base64, etc.
	CC->redirect_buffer = NewStrBufPlain(NULL, SIZ);
	CtdlOutputPreLoadedMsg(msg, MT_CITADEL, HEADERS_ALL, 0, 1, 0);
	CM_Free(msg);
	msgtext = CC->redirect_buffer;
	CC->redirect_buffer = NULL;
	if (msgtext != NULL) {
		syslog(LOG_DEBUG, "fulltext: wordbreaking message %ld (%d bytes)", msgnum, StrLength(msgtext));
	}
	txt = SmashStrBuf(&msgtext);
	tokens_in_this_message = wordbreaker(txt);
	free(txt);

	if (tokens_in_this_message == NULL) {
		syslog(LOG_DEBUG, "fulltext: skipping empty index for message %ld", msgnum);
		return;
	}

	syslog(LOG_DEBUG, "fulltext: %sindexing message %ld [%d tokens]",
		(op ? "" : "de"),
		msgnum,
		array_len(tokens_in_this_message)
	);

	if (array_len(tokens_in_this_message) > 0) {
		begin_critical_section(S_INDEXER);
		cdb_begin_transaction();
		for (i=0; i<array_len(tokens_in_this_message); ++i) {

			// Identify the bucket which we will be modifying
			memcpy(&tok, array_get_element_at(tokens_in_this_message, i), sizeof(int));
	
			// fetch the bucket
			cdb_bucket = cdb_fetch(CDB_FULLTEXT, &tok, sizeof(int));
			long *newbucket = malloc(cdb_bucket.len + sizeof(long));
			int nmsgs = cdb_bucket.len / sizeof(long);

			if (op == 1) {						// indexing, add this message to the bucket
				memcpy(&newbucket[0], cdb_bucket.ptr, cdb_bucket.len);
				int already_there = 0;
				for (j=0; j<nmsgs; ++j) {
					if (newbucket[j] == msgnum) {
						already_there = 1;
					}
				}
				if (already_there == 0) {
					memcpy(&newbucket[nmsgs++], &msgnum, sizeof(long));
				}
			}

			else if (op == 0) {					// deindexing, remove this message from the bucket
				memcpy(newbucket, cdb_bucket.ptr, cdb_bucket.len);
				for (j=0; j<nmsgs; ++j) {
					if ((newbucket[j] == msgnum) || (newbucket[j] == 0)) {
						memcpy(&newbucket[j], &newbucket[j+1], ((nmsgs-j)*sizeof(long)));
						--j;
						--nmsgs;
					}
				}
			}

			// Then write it back to disk
			cdb_store(CDB_FULLTEXT, &tok, sizeof(int), newbucket, (nmsgs*sizeof(long)));
			free(newbucket);

			if (server_shutting_down) break;
		}
		cdb_end_transaction();
		end_critical_section(S_INDEXER);
		CtdlSetConfigLong("MMfulltext", msgnum);
	}
	array_free(tokens_in_this_message);
}


// Scan a room for messages to index.
void ft_index_room(struct ctdlroom *qrbuf, void *data) {
	if (server_shutting_down) {
		return;
	}

	int num_msgs = 0;
	long *msglist;
	int i;

	// 2023aug30 ajc - old code did another CtdlGetRoom() here.  Not only is that redundant,
	// but for some reason it also made Berkeley DB deadlock after a while.  I don't know why.

	// qrbuf is already populated.  fetch the list of messages in this room.
	num_msgs = CtdlFetchMsgList(qrbuf->QRnumber, &msglist);

	// Identify messages which have NOT yet been seen by the indexer.
	if (msglist != NULL) {
		for (i=0; i<num_msgs; ++i) {
			if (
				(msglist[i] > 0)
				&& (msglist[i] > highest_msg_already_indexed)
				&& (msglist[i] <= highest_msg_to_be_indexed)
			   ) {
				array_append(messages_to_be_indexed, &msglist[i]);
			}
		}
		free(msglist);
	}
}


// Begin the fulltext indexing process.
void do_fulltext_indexing(void) {
	int i;
	static time_t last_progress = 0L;
	static int is_running = 0;

	if (is_running) return;         // Concurrency check - only one can run 
	is_running = 1;

	// Don't do this if the site doesn't have it enabled.
	if (!CtdlGetConfigInt("c_enable_fulltext")) {
		is_running = 0;
		return;
	}
	// If we've switched wordbreaker modules, burn the index and start over.
	begin_critical_section(S_CONTROL);
	if (CtdlGetConfigInt("MM_fulltext_wordbreaker") != FT_WORDBREAKER_ID) {
		syslog(LOG_DEBUG, "fulltext: wb ver on disk = %d, code ver = %d",
			CtdlGetConfigInt("MM_fulltext_wordbreaker"), FT_WORDBREAKER_ID
		);
		syslog(LOG_INFO, "fulltext: (re)initializing index");
		cdb_trunc(CDB_FULLTEXT);
		CtdlSetConfigLong("MMfulltext", 0);
	}
	end_critical_section(S_CONTROL);

	// Silently return if our fulltext index is up to date with new messages.
	if ((CtdlGetConfigLong("MMfulltext") >= CtdlGetConfigLong("MMhighest"))) {
		is_running = 0;
		return;		// nothing to do!
	}

	highest_msg_already_indexed = CtdlGetConfigLong("MMfulltext");
	highest_msg_to_be_indexed = CtdlGetConfigLong("MMhighest");
	syslog(LOG_DEBUG, "fulltext: indexing started. msgs %ld--%ld", highest_msg_already_indexed, highest_msg_to_be_indexed);

	messages_to_be_indexed = array_new(sizeof(long));

	// Now go through each room and find messages to index.
	CtdlForEachRoom(ft_index_room, NULL);				// load all msg pointers
	array_sort(messages_to_be_indexed, longcmp);			// sort them

	// Here it is ... do each message!
	long msgnum = 0;
	long prev_msgnum = 0;
	time_t started_indexing_at = time(NULL);
	int yielded = 0;
	for (i=0; ((i<array_len(messages_to_be_indexed)) && (yielded == 0)); ++i) {
		memcpy(&msgnum, array_get_element_at(messages_to_be_indexed, i), sizeof(long));

		if (msgnum != prev_msgnum) {				// careful to avoid dupes
			ft_index_message(msgnum, 1);
		}
		prev_msgnum = msgnum;
		
		// If we run too long, yield the thread so the server can do other things.  We'll be back.
		if (time(NULL) - started_indexing_at >= MAXIMUM_INDEXER_RUN_TIME) {
			yielded = 1;
		}
	}

	array_free(messages_to_be_indexed);
	CtdlSetConfigInt("MM_fulltext_wordbreaker", FT_WORDBREAKER_ID);
	syslog(LOG_DEBUG, "fulltext: indexer has run for %ld seconds; yielded=%d", time(NULL) - started_indexing_at, yielded);

	// This keeps the indexer from starting up over and over if the highest message in the list was deleted
	// before we had a chance to index it.
	if (!yielded) {
		CtdlSetConfigLong("MMfulltext", highest_msg_to_be_indexed);
	}

	is_running = 0;
	return;
}


// API call to perform searches.
// (This one does the "all of these words" search.)
// Caller is responsible for freeing the message list.
Array *CtdlFullTextSearch(const char *search_string) {
	int i, j, tok;
	struct cdbdata cdb_bucket;
	long msgnum, smsgnum;
	int count;

	Array *r = array_new(sizeof(long));
	if (!r) return(NULL);
	Array *t = wordbreaker(search_string);

	if ((t != NULL) && (array_len(t) > 0)) {
		for (i=0; i<array_len(t); ++i) {
			memcpy(&tok, array_get_element_at(t, i), sizeof(int));
			cdb_bucket = cdb_fetch(CDB_FULLTEXT, &tok, sizeof(int));
			if (cdb_bucket.ptr != NULL) {
				for (j=0; j<(cdb_bucket.len / sizeof(long)); ++j) {
					memcpy(&msgnum, cdb_bucket.ptr + (j*sizeof(long)), sizeof(long));
					array_append(r, &msgnum);
				}
			}
		}
	}

	// We need to return any message containing ALL of the tokens.
	// If a message number appears in the array `n` times, where `n` is the number of search words,
	// that means it matched all search words.
	array_sort(r, longcmp);
	for (i=0; i<array_len(r); ++i) {
		memcpy(&msgnum, array_get_element_at(r, i), sizeof(long));
		count = 1;
		for (j=i+1; j<array_len(r); ++j) {
			memcpy(&smsgnum, array_get_element_at(r, j), sizeof(long));
			if (msgnum == smsgnum) {
				++count;
				array_delete_element_at(r, j);
				--j;
			}
		}
		if (count != array_len(t)) {
			array_delete_element_at(r, i);
			--i;
		}
	}

	array_free(t);
	return(r);
}


// This search command is for diagnostic purposes and may be removed or replaced.
void cmd_srch(char *argbuf) {
	int i;
	char search_string[SIZ];
	Array *matches = NULL;
	long msgnum;

	if (CtdlAccessCheck(ac_logged_in)) return;

	if (!CtdlGetConfigInt("c_enable_fulltext")) {
		cprintf("%d Full text index is not enabled on this server.\n", ERROR + CMD_NOT_SUPPORTED);
		return;
	}

	extract_token(search_string, argbuf, 0, '|', sizeof search_string);
	matches = CtdlFullTextSearch(search_string);

	cprintf("%d %d msgs match all search words:\n", LISTING_FOLLOWS, array_len(matches));
	if ((matches != NULL) && (array_len(matches) > 0)) {
		for (i=0; i<array_len(matches); ++i) {
			memcpy(&msgnum, array_get_element_at(matches, i), sizeof(long));
			cprintf("%ld\n", msgnum);
		}
	}
	if (matches != NULL) {
		array_free(matches);
	}
	cprintf("000\n");
}


void ft_delete_remove(char *room, long msgnum) {
	if (room) return;
	
	// Remove from fulltext index
	if (CtdlGetConfigInt("c_enable_fulltext")) {
		ft_index_message(msgnum, 0);
	}
}


// Initialization function, called from modules_init.c
char *ctdl_module_init_fulltext(void) {
	if (!threading) {
		CtdlRegisterProtoHook(cmd_srch, "SRCH", "Full text search");
		CtdlRegisterDeleteHook(ft_delete_remove);
		CtdlRegisterSessionHook(do_fulltext_indexing, EVT_HOUSE, PRIO_CLEANUP + 300);
	}
	// return our module name for the log
	return "fulltext";
}
