// This module implements server commands related to the display and
// manipulation of the "Who's online" list.
//
// Copyright (c) 1987-2024 by the citadel.org team
//
// This program is open source software.  Use, duplication, or disclosure
// is subject to the terms of the GNU General Public License, version 3.

#include "../../sysdep.h"
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <pwd.h>
#include <errno.h>
#include <sys/types.h>
#include <time.h>
#include <sys/wait.h>
#include <string.h>
#include <limits.h>
#include <libcitadel.h>
#include "../../citadel_defs.h"
#include "../../server.h"
#include "../../citserver.h"
#include "../../support.h"
#include "../../config.h"
#include "../../control.h"
#include "../../user_ops.h"
#include "../../database.h"
#include "../../msgbase.h"
#include "../../ctdl_module.h"

// Don't show the names of private rooms unless the viewing
// user also knows the rooms.
void GenerateRoomDisplay(char *real_room, CitContext *viewed, CitContext *viewer) {
	int ra;

	strcpy(real_room, viewed->room.QRname);

	if (viewed->room.QRflags & QR_MAILBOX) {
		strcpy(real_room, &real_room[11]);
	}

	if (viewed->room.QRflags & QR_PRIVATE) {
		CtdlRoomAccess(&viewed->room, &viewer->user, &ra, NULL);
		if ( (ra & UA_KNOWN) == 0) {
			strcpy(real_room, " ");
		}
	}

	if (viewed->cs_flags & CS_CHAT) {
		while (strlen(real_room) < 14) {
			strcat(real_room, " ");
		}
		strcpy(&real_room[14], "<chat>");
	}

}


// display who's online
void cmd_rwho(char *argbuf) {
	struct CitContext *nptr;
	int nContexts, i;
	int aide;
	char room[ROOMNAMELEN];
	char flags[5];
	
	// So that we don't keep the context list locked for a long time
	// we create a copy of it first
	nptr = CtdlGetContextArray(&nContexts) ;
	if (!nptr) {
		// Couldn't malloc so we have to bail but stick to the protocol
		cprintf("%d%c \n", LISTING_FOLLOWS, CtdlCheckExpress(CC) );
		cprintf("000\n");
		return;
	}
	
	aide = ( (CC->user.axlevel >= AxAideU) || (CC->internal_pgm) ) ;
	cprintf("%d%c \n", LISTING_FOLLOWS, CtdlCheckExpress(CC) );
	
	for (i=0; i<nContexts; i++)  {
		flags[0] = '\0';
		
		if (!aide && nptr[i].state == CON_SYS)
			continue;

		if (!aide && nptr[i].kill_me != 0)
			continue;

		if (nptr[i].cs_flags & CS_POSTING) {
			strcat(flags, "*");
		}
		else {
			strcat(flags, ".");
		}
		   
		GenerateRoomDisplay(room, &nptr[i], CC);

		if ((nptr[i].cs_flags & CS_STEALTH) && (aide)) {
			strcat(flags, "-");
		}
		
		if (((nptr[i].cs_flags&CS_STEALTH)==0) || (aide)) {
			cprintf("%d|%s|%s|%s|%s|%ld|%s|%s||||%d|%d\n",
				nptr[i].cs_pid,
				(nptr[i].logged_in ? nptr[i].curr_user : NLI),
				room,
				nptr[i].cs_host,
				nptr[i].cs_clientname,
				(long)(nptr[i].lastidle),
				nptr[i].lastcmdname,
				flags,
				nptr[i].logged_in,
				nptr[i].state
			);
		}
	}
	
	// release our copy of the context list
	free(nptr);

	// Now it's magic time.  Before we finish, call any EVT_RWHO hooks
	// so that external paging modules such as serv_icq can add more
	// content to the Wholist.
	PerformSessionHooks(EVT_RWHO);
	cprintf("000\n");
}


// enter or exit "stealth mode"
void cmd_stel(char *cmdbuf) {
	int requested_mode;

	requested_mode = extract_int(cmdbuf,0);

	if (CtdlAccessCheck(ac_logged_in)) return;

	if (requested_mode == 1) {
		CC->cs_flags = CC->cs_flags | CS_STEALTH;
		PerformSessionHooks(EVT_STEALTH);
	}
	if (requested_mode == 0) {
		CC->cs_flags = CC->cs_flags & ~CS_STEALTH;
		PerformSessionHooks(EVT_UNSTEALTH);
	}

	cprintf("%d %d\n", CIT_OK,
		((CC->cs_flags & CS_STEALTH) ? 1 : 0) );
}


// Initialization function, called from modules_init.c
char *ctdl_module_init_rwho(void) {
	if (!threading) {
	        CtdlRegisterProtoHook(cmd_rwho, "RWHO", "Display who is online");
	        CtdlRegisterProtoHook(cmd_stel, "STEL", "Enter/exit stealth mode");

	}
	
	// return our module name for the log
        return "rwho";
}
