// Dump the Citadel database to a flat file that can be restored by ctdlload on any architecture
//
// Copyright (c) 2023-2024 by Art Cancro citadel.org
//
// This program is open source software.  Use, duplication, or disclosure
// is subject to the terms of the GNU General Public License, version 3.

#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <stdio.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netdb.h>
#include <string.h>
#include <pwd.h>
#include <errno.h>
#include <stdarg.h>
#include <limits.h>
#include <syslog.h>
#include <libcitadel.h>
#include <zlib.h>
#include "../server/sysdep.h"
#include "../server/citadel_defs.h"
#include "../server/server.h"
#include "../server/citadel_dirs.h"
#include "../server/database.h"

uid_t ctdluid = 0;

// Wrapper for realloc() that crashes and burns if the call fails.
void *reallok(void *ptr, size_t size) {
	void *p = realloc(ptr, size);
	if (!p) {
		fprintf(stderr, "realloc() failed to resize %p to %ld bytes, error: %m\n", ptr, size);
		abort();
	}
	return p;
}


// convert a binary blob to base64 (non-reentrant!)
char *b64out(void *data, size_t len) {
	static char *outbuf = NULL;
	static size_t outlen = 0;

	if ((outbuf == NULL) || (outlen < (len * 2))) {
		outbuf = reallok(outbuf, (len * 2));
		outlen = len * 2;
	}

	CtdlEncodeBase64(outbuf, data, len, 0);
	return(outbuf);
}


// export function for a message in msgmain
void export_msgmain(int which_cdb, struct cdbkeyval kv) {
	long in_msgnum;

	memcpy(&in_msgnum, kv.key.ptr, sizeof(in_msgnum));

	// If the msgnum is negative, we are looking at METADATA
	if (in_msgnum < 0) {
		struct MetaData *meta = (struct MetaData *)kv.val.ptr;
		printf("msgmeta|%ld|%d|%s|%ld|\n",
			meta->meta_msgnum,
			meta->meta_refcount,
			meta->meta_content_type,
			meta->meta_rfc822_length
		);
	}

	// If the msgnum is positive, we are looking at a MESSAGE
	else if (in_msgnum > 0) {
		printf("msgtext|%ld|%s|\n", in_msgnum, b64out(kv.val.ptr, kv.val.len));
	}

	// If the msgnum is 0 it's probably not a valid record.
}


// export function for a user record
void export_user(int which_cdb, struct cdbkeyval kv) {

	struct ctdluser *user = (struct ctdluser *)kv.val.ptr;

	printf("user|%d|%d|%s|%u|%d|%ld|%ld|%d|%s|%ld|%ld|%s|%ld|%ld|\n",
		user->version,
		user->uid,
		user->password,
		user->flags,
		user->axlevel,
		user->usernum,
		user->lastcall,
		user->USuserpurge,
		user->fullname,
		user->msgnum_bio,
		user->msgnum_pic,
		b64out(user->emailaddrs, strlen(user->emailaddrs)),
		user->msgnum_inboxrules,
		user->lastproc_inboxrules
	);
}


// export function for a room record
void export_room(int which_cdb, struct cdbkeyval kv) {

	struct ctdlroom *room = (struct ctdlroom *)kv.val.ptr;

	printf("room|%s|%s|%ld|%ld|%ld|%u|%s|%ld|%d|%ld|%d|%d|%ld|%d|%u|%d|%ld|\n",
		room->QRname,
		room->QRpasswd,
		room->QRroomaide,
		room->QRhighest,
		room->QRgen,
		room->QRflags,
		room->QRdirname,
		room->msgnum_info,
		room->QRfloor,
		room->QRmtime,
		room->QRep.expire_mode,
		room->QRep.expire_value,
		room->QRnumber,
		room->QRorder,
		room->QRflags2,
		room->QRdefaultview,
		room->msgnum_pic
	);
}


// export function for a floor record
void export_floor(int which_cdb, struct cdbkeyval kv) {

	int floor_num;
	memcpy(&floor_num, kv.key.ptr, sizeof(int));

	struct floor *floor = (struct floor *)kv.val.ptr;

	printf("floor|%d|%u|%s|%d|%d|%d|\n",
		floor_num,
		floor->f_flags,
		floor->f_name,
		floor->f_ref_count,
		floor->f_ep.expire_mode,
		floor->f_ep.expire_value
	);
}


// export function for a msglist
// (indexed by a long and the data is arrays of longs)
void export_msglist(int which_cdb, struct cdbkeyval kv) {
	int i;
	int num_msgs;
	long msg;

	// records are indexed by a single "long" and contains an array of zero or more "long"s
	long roomnum;
	memcpy(&roomnum, kv.key.ptr, sizeof(long));

	printf("msglist|%ld|", roomnum);

	if (kv.val.len > 0) {
		num_msgs = kv.val.len / sizeof(long);
		for (i=0; i<num_msgs; ++i) {
			memcpy(&msg, (kv.val.ptr + (i * sizeof(long))), sizeof(long));
			if (i != 0) {
				printf(",");
			}
			printf("%ld", msg);
		}
	}
	printf("|\n");
}


// export function for a full text search index record
// (indexed by an int and the data is arrays of longs)
void export_fulltext(int which_cdb, struct cdbkeyval kv) {
	int i;
	int num_msgs;
	long msg;

	// records are indexed by a single "int" and contains an array of zero or more "long"s
	int indexnum;
	memcpy(&indexnum, kv.key.ptr, sizeof(int));

	printf("fulltext|%d|", indexnum);

	if (kv.val.len > 0) {
		num_msgs = kv.val.len / sizeof(long);
		for (i=0; i<num_msgs; ++i) {
			memcpy(&msg, (kv.val.ptr + (i * sizeof(long))), sizeof(long));
			if (i != 0) {
				printf(",");
			}
			printf("%ld", msg);
		}
	}
	printf("|\n");
}


// export function for a visit record
void export_visit(int which_cdb, struct cdbkeyval kv) {
	struct visit *visit = (struct visit *)kv.val.ptr;
	int i, len;

	// If there is corrupt data in the "seen" array, cut that out before exporting
	len = strlen(visit->v_seen);
	for (i=0; i<len; ++i) {
		if (!isprint(visit->v_seen[i])) {
			visit->v_seen[i] = 0;
		}
	}

	// If there is corrupt data in the "answered" array, cut that out before exporting
	len = strlen(visit->v_answered);
	for (i=0; i<len; ++i) {
		if (!isprint(visit->v_answered[i])) {
			visit->v_answered[i] = 0;
		}
	}

	// output the record
	printf("visit|%ld|%ld|%ld|%ld|%u|%s|%s|%d|\n",
		visit->v_roomnum,
		visit->v_roomgen,
		visit->v_usernum,
		visit->v_lastseen,
		visit->v_flags,
		visit->v_seen,
		visit->v_answered,
		visit->v_view
	);
}


// export function for a directory record
void export_dir(int which_cdb, struct cdbkeyval kv) {
	printf("dir|");
	fwrite(kv.key.ptr, kv.key.len, 1, stdout);
	printf("|%s|\n", (char *)kv.val.ptr);
}


// export function for a use table record
void export_usetable(int which_cdb, struct cdbkeyval kv) {
	struct UseTable *u = (struct UseTable *)kv.val.ptr;
	printf("use|%d|%ld|\n", u->hash, u->timestamp);
}


// export function for large message texts
void export_bigmsg(int which_cdb, struct cdbkeyval kv) {
	long msgnum;

	memcpy(&msgnum, kv.key.ptr, sizeof(msgnum));
	printf("bigmsg|%ld|%s|\n", msgnum, b64out(kv.val.ptr, kv.val.len));
}


// export function for EUID Index records
void export_euidindex(int which_cdb, struct cdbkeyval kv) {

	// The structure of an euidindex record *key* is:
	// |----room_number----|----------EUID-------------|
	//    (sizeof long)       (actual length of euid)

	// The structure of an euidindex record *value* is:
	// |-----msg_number----|----room_number----|----------EUID-------------|
	//    (sizeof long)       (sizeof long)       (actual length of euid)

	long msgnum, roomnum;
	char *euid;

	memcpy(&msgnum, kv.val.ptr, sizeof(long));
	memcpy(&roomnum, kv.val.ptr+sizeof(long), sizeof(msgnum));
	euid = kv.val.ptr+(sizeof(long)*2);

	printf("euidindex|%ld|%ld|%s|\n", msgnum, roomnum, euid);
}


// export users-by-number records
// (This is a secondary index -- should we just regenerate the data after import?)
void export_usersbynumber(int which_cdb, struct cdbkeyval kv) {

	// key is a long
	long usernum;
	memcpy(&usernum, kv.key.ptr, sizeof(usernum));

	// value is a string
	printf("usersbynumber|%ld|%s|\n", usernum, (char *)kv.val.ptr);
}


// export function for a config record
void export_config(int which_cdb, struct cdbkeyval kv) {

	printf("config|%s|%s|\n",
		(char *)kv.val.ptr,
		(char *)kv.val.ptr + strlen(kv.val.ptr) + 1
	);

}


// For obsolete databases, zero all the output
void zero_function(int which_cdb, struct cdbkeyval kv) {
	// do nothing
}


void (*export_functions[])(int which_cdb, struct cdbkeyval kv) = {
	export_msgmain,		// 00 CDB_MSGMAIN	
	export_user,		// 01 CDB_USERS
	export_room,		// 02 CDB_ROOMS
	export_floor,		// 03 CDB_FLOORTAB
	export_msglist,		// 04 CDB_MSGLISTS
	export_visit,		// 05 CDB_VISIT
	export_dir,		// 06 CDB_DIRECTORY
	export_usetable,	// 07 CDB_USETABLE
	export_bigmsg,		// 08 CDB_BIGMSGS
	export_fulltext,	// 09 CDB_FULLTEXT
	export_euidindex,	// 0a CDB_EUIDINDEX
	export_usersbynumber,	// 0b CDB_USERSBYNUMBER
	zero_function,		// 0c CDB_UNUSED1 (obsolete)
	export_config		// 0d CDB_CONFIG
};


void export_table(int which_cdb) {
	int ret;
	struct cdbkeyval ckv;

	int num_good_rows = 0;
	int num_bad_rows = 0;

	cdb_rewind(which_cdb);
	while (ckv = cdb_next_item(which_cdb), ckv.val.ptr!=NULL) {		// always read through to the end
		// Call the export function registered to this table
		export_functions[which_cdb](which_cdb, ckv);
	}

	// Knowing the total number of rows isn't critical to the program.  It's just for the user to know.
	fflush(stdout);
}


int main(int argc, char **argv) {
	int i = 0;
	int confirmed = 0;
	char *ctdldir = CTDLDIR;

	// display the greeting
	fprintf(stderr,
		"\033[44m\033[1m╔════════════════════════════════════════════════════════════════════════╗\033[0m\n"
		"\033[44m\033[1m║ DB Dump utility for Citadel                               version %-4d ║\033[0m\n"
		"\033[44m\033[1m║ Copyright (c) 2023-2024 by citadel.org et al.                          ║\033[0m\n"
		"\033[44m\033[1m║ This program is open source software.  Use, duplication, or disclosure ║\033[0m\n"
		"\033[44m\033[1m║ is subject to the terms of the GNU General Public license v3.          ║\033[0m\n"
		"\033[44m\033[1m╚════════════════════════════════════════════════════════════════════════╝\033[0m\n",
		REV_LEVEL
	);

	// Parse command line
	int a;
	while ((a = getopt(argc, argv, "h:y")) != EOF) {
		switch (a) {
		case 'h':
			ctdldir = optarg;
			break;
		case 'y':
			confirmed = 1;
			break;
		default:
			fprintf(stderr, "%s: usage: %s -s citadel_dir [>dumpfile]\n", argv[0], argv[0]);
			exit(2);
		}
	}

	if (confirmed == 1) {
		fprintf(stderr, "ctdldump: You have specified the [-y] flag, so processing will continue.\n");
	}
	else {
		fprintf(stderr, "ctdldump: usage: ctdldump -y -h[citadel_dir] >[dump_file]\n");
		fprintf(stderr, "          -y : yes, I know this program can do damage and I want to run it anyway.\n");
		fprintf(stderr, "          -h : [citadel_dir] is your server directory, usually /usr/local/citadel\n");
		fprintf(stderr, "          Please read [ https://www.citadel.org/dump-and-load.html ] to learn how to proceed.\n");
		exit(1);
	}

	if (chdir(ctdldir) != 0) {
		fprintf(stderr, "ctdlload: unable to change directory to [%s]: %m", ctdldir);
		exit(2);
	}

	// backend modules use syslog -- redirect to stderr
	openlog("ctdldump", LOG_PERROR , LOG_DAEMON);

	// initialize the database backend
	cdb_init_backends();
	cdb_open_databases();

	printf("begin|\n");
	for (i = 0; i < MAXCDB; ++i) {
		export_table(i);
	}
	printf("end|\n");
	fflush(stdout);

	// close databases
	cdb_close_databases();

	fprintf(stderr, "ctdldump: \033[32m\033[1mfinished\033[0m\n");
	exit(0);
}
