// Load (restore) the Citadel database from a flat file created by ctdldump
//
// Copyright (c) 2023-2024 by Art Cancro citadel.org
//
// This program is open source software.  Use, duplication, or disclosure
// is subject to the terms of the GNU General Public License, version 3.

#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <stdio.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netdb.h>
#include <string.h>
#include <pwd.h>
#include <errno.h>
#include <stdarg.h>
#include <limits.h>
#include <syslog.h>
#include <libcitadel.h>
#include <zlib.h>
#include "../server/sysdep.h"
#include "../server/citadel_defs.h"
#include "../server/server.h"
#include "../server/makeuserkey.h"
#include "../server/citadel_dirs.h"
#include "../server/database.h"

uid_t ctdluid = 0;

// Wrapper for realloc() that crashes and burns if the call fails.
void *reallok(void *ptr, size_t size) {
	void *p = realloc(ptr, size);
	if (!p) {
		fprintf(stderr, "realloc() failed to resize %p to %ld bytes, error: %m\n", ptr, size);
		abort();
	}
	return p;
}


// Convert a "msgtext" record to a message on disk.   NOT THREADSAFE
// This also works for "bigmsg" records.
int import_msgtext(char *line, struct cdbkeyval *kv) {

	static char *b64_decoded_msg = NULL;
	static size_t b64_decoded_alloc = 0;
	long msgnum;
	char *token;

	token = strtok(line, "|");
	msgnum = atol(strtok(NULL, "|"));
	token = strtok(NULL, "|");

	// The record key will be the message number
	kv->key.len = sizeof(long);
	kv->key.ptr = reallok(kv->key.ptr, kv->key.len);
	memcpy(kv->key.ptr, &msgnum, kv->key.len);

	// The record data will be the decoded message text.
	// We are allocating more memory than we need, but BDB will only write the number of bytes we tell it to.
	kv->val.ptr = (char *) reallok(kv->val.ptr, strlen(token));
	kv->val.len = CtdlDecodeBase64(kv->val.ptr, token, strlen(token));
	return(1);
}


// Convert a "msgmeta" record to a message metadata record on disk.  NOT THREADSAFE
int import_msgmeta(char *line, struct cdbkeyval *kv) {
	char *token;
	struct MetaData *m;

	m = malloc(sizeof(struct MetaData));
	if (!m) {
		fprintf(stderr, "import_msgmeta: malloc: %s\n", strerror(errno));
		return(0);
	}
	memset(m, 0, sizeof(struct MetaData));
	char *p = line;

	for (int i=0; (token = strsep(&p, "|")); ++i) {
		switch(i) {
			case 1:
				m->meta_msgnum = atol(token);
				break;
			case 2:
				m->meta_refcount = atoi(token);
				break;
			case 3:
				strncpy(m->meta_content_type, token, sizeof(m->meta_content_type));
				break;
			case 4:
				m->meta_rfc822_length = atol(token);
				break;
		}
	}

	// metadata records are stored in the CDB_MSGMAIN table,
	// but with the index being the *negative* of the message number.
	long index = 0 - m->meta_msgnum;
	kv->key.len = sizeof(long);
	kv->key.ptr = reallok(NULL, kv->key.len);
	memcpy(kv->key.ptr, &index, kv->key.len);

	// data
	kv->val.len = sizeof(struct MetaData);
	kv->val.ptr = (char *) m;				// out_data owns this memory now

	return(1);
}


// Convert a "user" record to a record on disk.  (Source string is unusable after this function is called.)
int import_user(char *line, struct cdbkeyval *kv) {
	char userkey[USERNAME_SIZE];
	char *token;
	struct ctdluser *u;

	u = malloc(sizeof(struct ctdluser));
	if (!u) {
		fprintf(stderr, "malloc failed\n");
		return(0);
	}
	memset(u, 0, sizeof(struct ctdluser));
	char *p = line;

	for (int i=0; (token = strsep(&p, "|")); ++i) {
		switch(i) {
			case 1:
				u->version = atoi(token);
				break;
			case 2:
				u->uid = atoi(token);
				break;
			case 3:
				safestrncpy(u->password, token, sizeof(u->password));
				break;
			case 4:
				u->flags = atoi(token);
				break;
			case 5:
				u->axlevel = atoi(token);
				break;
			case 6:
				u->usernum = atol(token);
				break;
			case 7:
				u->lastcall = atol(token);
				break;
			case 8:
				u->USuserpurge = atoi(token);
				break;
			case 9:
				safestrncpy(u->fullname, token, sizeof(u->fullname));
				break;
			case 10:
				u->msgnum_bio = atol(token);
				break;
			case 11:
				u->msgnum_pic = atol(token);
				break;
			case 12:
				CtdlDecodeBase64(token, token, strlen(token));			// Decode in place
				safestrncpy(u->emailaddrs, token, sizeof(u->emailaddrs));
				break;
			case 13:
				u->msgnum_inboxrules = atol(token);
				break;
			case 14:
				u->lastproc_inboxrules = atol(token);
				break;
		}
	}

	// reject any user records without names
	if (strlen(u->fullname) == 0) {
		free(u);
		return(0);
	}
	
	// ok, good to go
	makeuserkey(userkey, u->fullname);
	kv->key.len = strlen(userkey);
	kv->key.ptr = strdup(userkey);
	kv->val.len = sizeof(struct ctdluser);
	kv->val.ptr = (char *) u;
	return(1);
}


// Convert a "room" record to a record on disk.  (Source string is unusable after this function is called.)
int import_room(char *line, struct cdbkeyval *kv) {
	char *token;
	struct ctdlroom *r;

	r = malloc(sizeof(struct ctdlroom));
	if (!r) {
		fprintf(stderr, "import_room: malloc: %s\n", strerror(errno));
		return(0);
	}
	memset(r, 0, sizeof(struct ctdlroom));
	char *p = line;

	for (int i=0; (token = strsep(&p, "|")); ++i) {
		switch(i) {
			case 1:
 				strncpy(r->QRname, token, sizeof(r->QRname));
				break;
			case 2:
 				strncpy(r->QRpasswd, token, sizeof (r->QRpasswd));
				break;
			case 3:
 				r->QRroomaide = atol(token);
				break;
			case 4:
 				r->QRhighest = atol(token);
				break;
			case 5:
 				r->QRgen = atol(token);
				break;
			case 6:
 				r->QRflags = atoi(token);
				break;
			case 7:
 				strncpy(r->QRdirname, token, sizeof(r->QRdirname));
				break;
			case 8:
 				r->msgnum_info = atol(token);
				break;
			case 9:
 				r->QRfloor = atoi(token);
				break;
 			case 10:
				r->QRmtime = atol(token);
				break;
			case 11:
                		r->QRep.expire_mode = atoi(token);
				break;
			case 12:
                		r->QRep.expire_value = atoi(token);
				break;
			case 13:
 				r->QRnumber = atol(token);
				break;
			case 14:
 				r->QRorder = atoi(token);
				break;
			case 15:
 				r->QRflags2 = atoi(token);
				break;
			case 16:
 				r->QRdefaultview = atoi(token);
				break;
			case 17:
				r->msgnum_pic = atol(token);
				break;
		}
	}

	// The key is the room name in all lower case
	kv->key.len = strlen(r->QRname);
	kv->key.ptr = strdup(r->QRname);
	char *k = (char *)kv->key.ptr;
	for (int i=0; i<=kv->key.len; ++i) {
		k[i] = tolower(k[i]);
	}

	kv->val.len = sizeof(struct ctdlroom);
	kv->val.ptr = (char *) r;
	return(1);
}


// Convert a floor record to a record on disk.
int import_floor(char *line, struct cdbkeyval *kv) {
	char *token;
	struct floor *f;
	int floor_num;

	f = malloc(sizeof(struct floor));
	if (!f) {
		fprintf(stderr, "import_floor: malloc: %s\n", strerror(errno));
		return(0);
	}
	memset(f, 0, sizeof(struct floor));
	char *p = line;

	for (int i=0; (token = strsep(&p, "|")); ++i) {
		switch(i) {
			case 1:
				floor_num = atoi(token);
				break;
			case 2:
				f->f_flags = atoi(token);
				break;
			case 3:
				strncpy(f->f_name, token, sizeof(f->f_name));
				break;
			case 4:
				f->f_ref_count = atoi(token);
				break;
			case 5:
                		f->f_ep.expire_mode = atoi(token);
				break;
			case 6:
                		f->f_ep.expire_value = atoi(token);
				break;
		}
	}

	kv->key.len = sizeof(int);
	kv->key.ptr = malloc(kv->key.len);
	memcpy(kv->key.ptr, &floor_num, kv->key.len);

	kv->val.len = sizeof(struct floor);
	kv->val.ptr = (char *) f;
	return(1);
}


// Import a msglist record
// msglist|26|32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,50,51|
int import_msglist(char *line, struct cdbkeyval *kv) {
	long roomnum;
	char *token, *mtoken;
	char *p = line;
	char *q = NULL;
	int num_msgs = 0;
	long *msglist = NULL;

	for (int i=0; (token = strsep(&p, "|")); ++i) {
		switch(i) {
			case 1:
				roomnum = atol(token);
				break;
			case 2:
				q = token;
				for (int j=0; (mtoken = strsep(&q, ",")); ++j) {
					msglist = realloc(msglist, (num_msgs+1) * sizeof(long));
					msglist[num_msgs++] = atol(mtoken);
				}
				break;
		}
	}

	kv->key.len = sizeof(long);
	kv->key.ptr = malloc(kv->key.len);
	memcpy(kv->key.ptr, &roomnum, kv->key.len);

	kv->val.len = num_msgs * sizeof(long);
	kv->val.ptr = (char *) msglist;

	return(1);
}


// Convert a "visit" record to a record on disk.
int import_visit(char *line, struct cdbkeyval *kv) {
	char *token;
	struct visit *v;

	v = malloc(sizeof(struct visit));
	if (!v) {
		fprintf(stderr, "import_visit: malloc: %s\n", strerror(errno));
		return(0);
	}
	memset(v, 0, sizeof(struct visit));
	char *p = line;

	for (int i=0; (token = strsep(&p, "|")); ++i) {
		switch(i) {
			case 1:
				v->v_roomnum = atol(token);
				break;
			case 2:
				v->v_roomgen = atol(token);
				break;
			case 3:
				v->v_usernum = atol(token);
				break;
			case 4:
				v->v_lastseen = atoi(token);
				break;
			case 5:
				v->v_flags = atoi(token);
				break;
			case 6:
				strncpy(v->v_seen, token, sizeof(v->v_seen));
				break;
			case 7:
				strncpy(v->v_answered, token, sizeof(v->v_answered));
				break;
			case 8:
				v->v_view = atoi(token);
				break;
		}
	}

	// The key is the same as the first three data members (3 x long)
	kv->key.len = sizeof(long) * 3;
	kv->key.ptr = reallok(NULL, kv->key.len);
	memcpy(kv->key.ptr, v, kv->key.len);

	kv->val.len = sizeof(struct visit);
	kv->val.ptr = (char *) v;
	return(1);
}


// Convert a "dir" record to a record on disk.
int import_dir(char *line, struct cdbkeyval *kv) {
	char dirkey[SIZ];
	char username[USERNAME_SIZE];
	char *token;

	char *p = line;
	for (int i=0; (token = strsep(&p, "|")); ++i) {
		switch(i) {
			case 1:
				strncpy(dirkey, token, sizeof(dirkey));
				break;
			case 2:
				strncpy(username, token, sizeof(username));
				break;
		}
	}

	kv->key.len = strlen(dirkey);
	kv->key.ptr = reallok(NULL, kv->key.len);
	memcpy(kv->key.ptr, dirkey, strlen(dirkey));

	kv->val.len = strlen(username) + 1;
	kv->val.ptr = strdup(username);

	return(1);
}


// Convert a "usetable" record to a record on disk.
int import_usetable(char *line, struct cdbkeyval *kv) {
	char *token;
	struct UseTable *u;

	u = malloc(sizeof(struct UseTable));
	if (!u) {
		fprintf(stderr, "import_usetable: malloc: %s\n", strerror(errno));
		return(0);
	}
	memset(u, 0, sizeof(struct UseTable));
	char *p = line;

	for (int i=0; (token = strsep(&p, "|")); ++i) {
		switch(i) {
			case 1:
				u->hash = atoi(token);
				break;
			case 2:
				u->timestamp = atol(token);
				break;
		}
	}

	// the key is just an int (the hash)
	kv->key.len = sizeof(int);
	kv->key.ptr = reallok(NULL, kv->key.len);
	memcpy(kv->key.ptr, &u->hash, kv->key.len);

	kv->val.len = sizeof(struct UseTable);
	kv->val.ptr = (char *) u;
	return(1);
}


// Import a full text search index record.
// It's just like a msglists record: a key and a list of message numbers, but the key is "int" instead of "long"
int import_fulltext(char *line, struct cdbkeyval *kv) {
	int indexnum;
	char *token, *mtoken;
	char *p = line;
	char *q = NULL;
	int num_msgs = 0;
	long *msglist = NULL;

	for (int i=0; (token = strsep(&p, "|")); ++i) {
		switch(i) {
			case 1:
				indexnum = atoi(token);
				break;
			case 2:
				q = token;
				for (int j=0; (mtoken = strsep(&q, ",")); ++j) {
					msglist = realloc(msglist, (num_msgs+1) * sizeof(long));
					msglist[num_msgs++] = atol(mtoken);
				}
				break;
		}
	}

	kv->key.len = sizeof(int);
	kv->key.ptr = malloc(kv->key.len);
	memcpy(kv->key.ptr, &indexnum, kv->key.len);

	kv->val.len = num_msgs * sizeof(long);
	kv->val.ptr = (char *) msglist;

	return(1);
}


// Import an EUID Index record
// euidindex|msgnum|roomnum|euid|
int import_euidindex(char *line, struct cdbkeyval *kv) {
	char euid[SIZ];
	long msgnum;
	long roomnum;
	char *token;

	char *p = line;
	for (int i=0; (token = strsep(&p, "|")); ++i) {
		switch(i) {
			case 1:
				msgnum = atol(token);
				break;
			case 2:
				roomnum = atol(token);
				break;
			case 3:
				strncpy(euid, token, sizeof(euid));
				break;
		}
	}

	// The structure of an euidindex record *key* is:
	// |----room_number----|----------EUID-------------|
	//    (sizeof long)       (actual length of euid)
	kv->key.len = sizeof(long) + strlen(euid) + 1;
	kv->key.ptr = reallok(NULL, kv->key.len);
	memcpy(kv->key.ptr, &roomnum, sizeof(long));
	strcpy(kv->key.ptr + sizeof(long), euid);

	// The structure of an euidindex record *value* is:
	// |-----msg_number----|----room_number----|----------EUID-------------|
	//    (sizeof long)       (sizeof long)       (actual length of euid)
	kv->val.len = sizeof(long) + sizeof(long) + strlen(euid) + 1;
	kv->val.ptr = (char *) reallok(NULL, kv->val.len);
	memcpy(kv->val.ptr, &msgnum, sizeof(long));
	memcpy(kv->val.ptr + sizeof(long), &roomnum, sizeof(long));
	strcpy(kv->val.ptr + sizeof(long) + sizeof(long), euid);

	return(1);
}


// Import a "users by number" (secondary index) record
// The key is a "long"
int import_usersbynumber(char *line, struct cdbkeyval *kv) {
	char *token;
	long usernum;

	char *p = line;
	for (int i=0; (token = strsep(&p, "|")); ++i) {
		switch(i) {
			case 1:
				usernum = atol(token);
				break;
			case 2:
				kv->key.len = sizeof(long);
				kv->key.ptr = reallok(NULL, sizeof(long));
				memcpy(kv->key.ptr, &usernum, kv->key.len);
				kv->val.ptr = (char *) strdup(token);
				kv->val.len = strlen(kv->val.ptr) + 1;
				return(1);
		}
	}

	return(0);		// should never get here unless it's a bad record
}


// Import a config record
// The key is the config key
// The data is the config key, a null, the value, and another null
int import_config(char *line, struct cdbkeyval *kv) {
	char *token;
	char *p = line;
	char *k, *v;

	for (int i=0; (token = strsep(&p, "|")); ++i) {
		switch(i) {
			case 1:
				k = token;
				kv->key.len = strlen(token);
				kv->key.ptr = strdup(token);
				break;
			case 2:
				v = token;
				kv->val.len = strlen(k) + strlen(v) + 2;
				kv->val.ptr = (char *) reallok(NULL, kv->val.len);
				strcpy(kv->val.ptr, k);
				strcpy(kv->val.ptr + strlen(k) + 1, v);
				break;
		}
	}

	return(1);
}


// Ingest one line of dump data.  NOT REENTRANT
void ingest_one(char *line, struct cdbkeyval *kv) {

	static int good_rows = 0;
	static int bad_rows = 0;
	static int current_cdb = -1 ;
	char record_type[32];
	int row_was_good;

	// Clear out our record buffer
	memset(kv, 0, sizeof(struct cdbkeyval));
	row_was_good = 0;

	// Identify the record type we are currently working with,
	// then call the correct conversion function to load up our record buffer.
	extract_token(record_type, line, 0, '|', sizeof record_type);
	if (!strcasecmp(record_type, "msgtext")) {
		current_cdb = CDB_MSGMAIN;
		row_was_good = import_msgtext(line, kv);
	}
	else if (!strcasecmp(record_type, "msgmeta")) {
		current_cdb = CDB_MSGMAIN;
		row_was_good = import_msgmeta(line, kv);
	}
	else if (!strcasecmp(record_type, "user")) {
		current_cdb = CDB_USERS;
		row_was_good = import_user(line, kv);
	}
	else if (!strcasecmp(record_type, "room")) {
		current_cdb = CDB_ROOMS;
		row_was_good = import_room(line, kv);
	}
	else if (!strcasecmp(record_type, "floor")) {
		current_cdb = CDB_FLOORTAB;
		row_was_good = import_floor(line, kv);
	}
	else if (!strcasecmp(record_type, "msglist")) {
		current_cdb = CDB_MSGLISTS;
		row_was_good = import_msglist(line, kv);
	}
	else if (!strcasecmp(record_type, "visit")) {
		current_cdb = CDB_VISIT;
		row_was_good = import_visit(line, kv);
	}
	else if (!strcasecmp(record_type, "dir")) {
		current_cdb = CDB_DIRECTORY;
		row_was_good = import_dir(line, kv);
	}
	else if (!strcasecmp(record_type, "use")) {
		current_cdb = CDB_USETABLE;
		row_was_good = import_usetable(line, kv);
	}
	else if (!strcasecmp(record_type, "bigmsg")) {
		current_cdb = CDB_BIGMSGS;
		row_was_good = import_msgtext(line, kv);
	}
	else if (!strcasecmp(record_type, "fulltext")) {
		current_cdb = CDB_FULLTEXT;
		row_was_good = import_fulltext(line, kv);
	}
	else if (!strcasecmp(record_type, "euidindex")) {
		current_cdb = CDB_EUIDINDEX;
		row_was_good = import_euidindex(line, kv);
	}
	else if (!strcasecmp(record_type, "usersbynumber")) {
		current_cdb = CDB_USERSBYNUMBER;
		row_was_good = import_usersbynumber(line, kv);
	}
	else if (!strcasecmp(record_type, "config")) {
		current_cdb = CDB_CONFIG;
		row_was_good = import_config(line, kv);
	}
	else {
		current_cdb = -1 ;
	}

	// If the conversion function was successful, write the record to the database.
	if ( (row_was_good) && (current_cdb >= 0) ) {
		cdb_store(current_cdb, kv->key.ptr, kv->key.len, kv->val.ptr, kv->val.len);
		++good_rows;
	}
	else {
		++bad_rows;
	}

	fprintf(stderr, "\r   \033[32m%9d\033[0m / \033[31m%8d\033[0m ", good_rows, bad_rows);
	fflush(stderr);

}


// This is the loop that loads the dump data.  NOT REENTRANT
void ingest(void) {
	static size_t line_alloc = 1;
	static char *line;
	static size_t line_len = 0;
	int ch;
	int begin_found = 0;
	int end_found = 0;
	struct cdbkeyval kv;

	memset(&kv, 0, sizeof(struct cdbkeyval));

	line = reallok(NULL, line_alloc);

	do {
		line_len = 0;

		while (ch = getc(stdin), ((ch != '\n') && (ch > 0))) {
			if ((line_len+2) > line_alloc) {
				line_alloc *= 2;
				line = reallok(line, line_alloc);
			}
			line[line_len++] = ch;
			line[line_len] = 0;
		}
	
		if (ch <= 0) {
			return;
		}

		if (line_len > 0) {
			if ( (begin_found) && (!end_found) ) {
				ingest_one(line, &kv);
			}
			if (!strncasecmp(line, HKEY("begin|"))) {
				begin_found = 1;
				fprintf(stderr, "   good rows / bad rows:\n");
			}
			if (!strncasecmp(line, HKEY("end|"))) {
				fprintf(stderr, "\n");
				end_found = 1;
			}
		}

	} while (ch >= 0);

	if (!begin_found) {
		fprintf(stderr, "\033[31mWARNING: \"begin\" line was not found in the loaded data.\033[0m\n");
	}
	if (!end_found) {
		fprintf(stderr, "\033[31mWARNING: \"end\" line was not found in the loaded data.\033[0m\n");
	}
}


// Main entry point
int main(int argc, char **argv) {
	int confirmed = 0;
	char *ctdldir = CTDLDIR;

	// display the greeting
	fprintf(stderr,
		"\033[44m\033[1m╔════════════════════════════════════════════════════════════════════════╗\033[0m\n"
		"\033[44m\033[1m║ DB Load utility for Citadel                               version %-4d ║\033[0m\n"
		"\033[44m\033[1m║ Copyright (c) 2023-2024 by citadel.org et al.                          ║\033[0m\n"
		"\033[44m\033[1m║ This program is open source software.  Use, duplication, or disclosure ║\033[0m\n"
		"\033[44m\033[1m║ is subject to the terms of the GNU General Public license v3.          ║\033[0m\n"
		"\033[44m\033[1m╚════════════════════════════════════════════════════════════════════════╝\033[0m\n",
		REV_LEVEL
	);

	// Parse command line
	int a;
	while ((a = getopt(argc, argv, "h:y")) != EOF) {
		switch (a) {
		case 'h':
			ctdldir = optarg;
			break;
		case 'y':
			confirmed = 1;
			break;
		default:
			fprintf(stderr, "%s: usage: %s -h citadel_dir [<dumpfile]\n", argv[0], argv[0]);
			exit(2);
		}
	}

	if (confirmed == 1) {
		fprintf(stderr, "ctdlload: You have specified the [-y] flag, so processing will continue.\n");
	}
	else {
		fprintf(stderr, "ctdlload: usage: ctdlload -y -h[citadel_dir] <[dump_file]\n");
		fprintf(stderr, "          -y : yes, I know this program can do damage and I want to run it anyway.\n");
		fprintf(stderr, "          -h : [citadel_dir] is your server directory, usually /usr/local/citadel\n");
		fprintf(stderr, "          Please read [ https://www.citadel.org/dump-and-load.html ] to learn how to proceed.\n");
		exit(1);
	}

	if (chdir(ctdldir) != 0) {
		fprintf(stderr, "ctdlload: unable to change directory to [%s]: %m", ctdldir);
		exit(2);
	}

	// backend modules use syslog -- redirect to stderr
	openlog("ctdlload", LOG_PERROR , LOG_DAEMON);

	// Remove any database that is already in the target directory (yes, delete it, be careful)
	system("rm -fv ./data/*");

	// initialize the database backend
	cdb_init_backends();
	cdb_open_databases();

	// Load rows from the input source
	ingest();

	// close databases
	cdb_close_databases();

	fprintf(stderr, "ctdlload: \033[32m\033[1mfinished\033[0m\n");
	exit(0);
}
