// Check to see if two events overlap.
//
// This function is used in Citadel Server and in both generations of WebCit.  Symlink it accordingly.
//
// Copyright (c) 1987-2024 by the citadel.org team
//
// This program is open source software.  Use, duplication, or disclosure is subject to the GNU General Public License version 3.

#include <libical/ical.h>
#include <string.h>
#include <syslog.h>

// Check to see if two events overlap.  Returns nonzero if they do.
int ical_ctdl_is_overlap(
	struct icaltimetype t1start,
	struct icaltimetype t1end,
	struct icaltimetype t2start,
	struct icaltimetype t2end
) {
	if (icaltime_is_null_time(t1start)) return(0);
	if (icaltime_is_null_time(t2start)) return(0);

	// if either event lacks end time, assume end = start
	if (icaltime_is_null_time(t1end)) {
		memcpy(&t1end, &t1start, sizeof(struct icaltimetype));
	}
	else {
		if (t1end.is_date && icaltime_compare(t1start, t1end)) {
			// The end date is non-inclusive so adjust it by one day because our test is inclusive.
			// Note that a day is not too much because we are talking about all day events.
			// If start==end we assume that it means one day.
			icaltime_adjust(&t1end, -1, 0, 0, 0);	
		}
	}

#ifdef DEBUG_ICAL_OVERLAP
	syslog(LOG_DEBUG, "Comparing t1start=%s(%s),t1end=%s(%s) to t2start=%s(%s),t2end=%s(%s)",
		icaltime_as_ical_string_r(t1start),	icaltime_get_tzid(t1start),
		icaltime_as_ical_string_r(t1end),	icaltime_get_tzid(t1end),
		icaltime_as_ical_string_r(t2start),	icaltime_get_tzid(t2start),
		icaltime_as_ical_string_r(t2end),	icaltime_get_tzid(t2end)
	);
#endif

	if (icaltime_is_null_time(t2end)) {
		memcpy(&t2end, &t2start, sizeof(struct icaltimetype));
	}
	else {
		if (t2end.is_date && icaltime_compare(t2start, t2end)) {
			icaltime_adjust(&t2end, -1, 0, 0, 0);	
		}
	}

	// First, check for all-day events
	if (t1start.is_date || t2start.is_date) {
		// If event 1 ends before event 2 starts, there is no match.
		if (icaltime_compare_date_only(t1end, t2start) < 0) return(0);

		// If event 2 ends before event 1 starts, there is no match.
		if (icaltime_compare_date_only(t2end, t1start) < 0) return(0);

		// Otherwise there is at least one day of overlap, so yes we have a match.
		return(1);
	}

	// Now check for overlaps using date *and* time.
	// We could probably implement this in terms of icaltime_span_overlaps() but this does the same thing.

	// If event 1 ends before event 2 starts, there is no match.
	if (icaltime_compare(t1end, t2start) <= 0) return(0);

	// If event 2 ends before event 1 starts, there is no match.
	if (icaltime_compare(t2end, t1start) <= 0) return(0);

	// Otherwise, there is an overlap, so yes we have a match.
	return(1);
}
